Hi, here's the amended patch.
Cheers,
 Ondrej
>From 29d3fbb14e59c883750cde3468fdc1bab5a227e1 Mon Sep 17 00:00:00 2001
From: Ondrej Oprala <[email protected]>
Date: Thu, 9 Aug 2012 17:34:09 +0200
Subject: [PATCH] du: Fix an issue with bogus warnings on bind-mounted
 directories

* NEWS: Mention the fix.
* src/du.c: Include "mountlist.h".
  (di_mnt): New global set.
  (fill_mount_table): New function.
  (hash_ins): Add HT parameter.
  (process_file): Look up each dir dev/ino pair in the new set.
  (main): Allocate, initialize, and free the new set.
* tests/Makefile.am: Add a new file.
* tests/du/bind-mount-dir-cycle: Add a test for the fix.
---
 NEWS              |  3 +++
 src/du.c          | 61 ++++++++++++++++++++++++++++++++++++++++++++++++++++---
 tests/Makefile.am |  1 +
 3 files changed, 62 insertions(+), 3 deletions(-)

diff --git a/NEWS b/NEWS
index ca4568a..8f85c2b 100644
--- a/NEWS
+++ b/NEWS
@@ -4,6 +4,9 @@ GNU coreutils NEWS                                    -*- 
outline -*-
 
 ** Bug fixes
 
+  du no longer emits bogus warnings when traversing bind-mounted
+  directory cycles.
+
   cksum now prints checksums atomically so that concurrent
   processes will not intersperse their output.
   [the bug dates back to the initial implementation]
diff --git a/tests/Makefile.am b/tests/Makefile.am
index edc04b4..ebaceb6 100644
--- a/tests/Makefile.am
+++ b/tests/Makefile.am
@@ -32,6 +32,7 @@ root_tests =                                  \
   cp/sparse-fiemap                             \
   dd/skip-seek-past-dev                                \
   df/problematic-chars                         \
+  du/bind-mount-dir-cycle                      \
   install/install-C-root                       \
   ls/capability                                        \
   ls/nameless-uid                              \
diff --git a/src/du.c b/src/du.c
index 7333941..bfe543e 100644
--- a/src/du.c
+++ b/src/du.c
@@ -35,6 +35,7 @@
 #include "exclude.h"
 #include "fprintftime.h"
 #include "human.h"
+#include "mountlist.h"
 #include "quote.h"
 #include "quotearg.h"
 #include "stat-size.h"
@@ -63,6 +64,9 @@ extern bool fts_debug;
 /* A set of dev/ino pairs.  */
 static struct di_set *di_set;
 
+/* A hash table for mount points.  */
+static struct di_set *di_mnt;
+
 /* Keep track of the preceding "level" (depth in hierarchy)
    from one call of process_file to the next.  */
 static size_t prev_level;
@@ -333,11 +337,11 @@ Mandatory arguments to long options are mandatory for 
short options too.\n\
   exit (status);
 }
 
 /* Try to insert the INO/DEV pair into the global table, HTAB.
    Return true if the pair is successfully inserted,
    false if the pair is already in the table.  */
 static bool
-hash_ins (ino_t ino, dev_t dev)
+hash_ins (struct di_set *di_set, ino_t ino, dev_t dev)
 {
   int inserted = di_set_insert (di_set, dev, ino);
   if (inserted < 0)
@@ -461,7 +465,7 @@ process_file (FTS *fts, FTSENT *ent)
       if (excluded
           || (! opt_count_all
               && (hash_all || (! S_ISDIR (sb->st_mode) && 1 < sb->st_nlink))
-              && ! hash_ins (sb->st_ino, sb->st_dev)))
+              && ! hash_ins (di_set, sb->st_ino, sb->st_dev)))
         {
           /* If ignoring a directory in preorder, skip its children.
              Ignore the next fts_read output too, as it's a postorder
@@ -476,6 +480,17 @@ process_file (FTS *fts, FTSENT *ent)
           return true;
         }
 
+      /*Check if dir is already in the mount table.  */
+      if (S_ISDIR (sb->st_mode))
+        {
+          if (di_set_lookup (di_mnt, sb->st_dev, sb->st_ino))
+            {
+              fts_set (fts, ent, FTS_SKIP);
+              error (0, 0, _("mount point %s already traversed"), quote 
(file));
+              return false;
+            }
+        }
+
       switch (info)
         {
         case FTS_D:
@@ -623,6 +638,38 @@ du_files (char **files, int bit_flags)
   return ok;
 }
 
+/* Fill the di_mnt table with dev/ino pairs of mount points.  */
+
+static void
+fill_mount_table (void)
+{
+  struct mount_entry *mnt_ent;
+  struct mount_entry *mnt_free;
+  struct stat *buf = xmalloc (sizeof *buf);
+
+  mnt_ent = read_file_system_list (false);
+  while (mnt_ent)
+    {
+      if (!mnt_ent->me_remote && !mnt_ent->me_dummy)
+        {
+          if (!stat (mnt_ent->me_mountdir, buf))
+            hash_ins (di_mnt, buf->st_ino, buf->st_dev);
+          else
+            error (0, errno, _("warning: unable to stat %s"),
+                   mnt_ent->me_mountdir);
+        }
+
+      mnt_free = mnt_ent;
+      mnt_ent = mnt_ent->me_next;
+
+      free (mnt_free->me_devname);
+      free (mnt_free->me_mountdir);
+      free (mnt_free->me_type);
+      free (mnt_free);
+    }
+  free (buf);
+}
+
 int
 main (int argc, char **argv)
 {
@@ -922,6 +969,13 @@ main (int argc, char **argv)
     xalloc_die ();
 
   /* Initialize the set of dev,inode pairs.  */
+
+  di_mnt = di_set_alloc ();
+  if (!di_mnt)
+    xalloc_die ();
+
+  fill_mount_table ();
+
   di_set = di_set_alloc ();
   if (!di_set)
     xalloc_die ();
@@ -1002,6 +1056,7 @@ main (int argc, char **argv)
 
   argv_iter_free (ai);
   di_set_free (di_set);
+  di_set_free (di_mnt);
 
   if (files_from && (ferror (stdin) || fclose (stdin) != 0) && ok)
     error (EXIT_FAILURE, 0, _("error reading %s"), quote (files_from));
@@ -333,11 +337,11 @@ Mandatory arguments to long options are mandatory for 
 }
 
-/* Try to insert the INO/DEV pair into the global table, HTAB.
+/* Try to insert the INO/DEV pair into the di_set table.
    Return true if the pair is successfully inserted,
    false if the pair is already in the table.  */
-- 
1.7.11.2

Reply via email to